@@ -1,23 +1,27 @@
-# -*- coding: utf-8 -*-
 """Implements JSON version of xonsh history backend."""
-import os
-import sys
-import time
 import collections
-import threading
 import collections.abc as cabc
+import os
+import re
+import sys
+import threading
+import time
 
 from xonsh.built_ins import XSH
 
 try:
     import ujson as json
+
+    JSONDecodeError = json.JSONDecodeError  # type: ignore
 except ImportError:
     import json  # type: ignore
 
-from xonsh.history.base import History
-import xonsh.tools as xt
+    JSONDecodeError = json.decoder.JSONDecodeError  # type: ignore
+
 import xonsh.lazyjson as xlj
+import xonsh.tools as xt
 import xonsh.xoreutils.uptime as uptime
+from xonsh.history.base import History
 
 
 def _xhj_gc_commands_to_rmfiles(hsize, files):
@@ -165,7 +169,7 @@ class JsonHistoryGC(threading.Thread):
         files = self.files(only_unlocked=True)
         rmfiles_fn = self.gc_units_to_rmfiles.get(units)
         if rmfiles_fn is None:
-            raise ValueError("Units type {0!r} not understood".format(units))
+            raise ValueError(f"Units type {units!r} not understood")
 
         size_over, rm_files = rmfiles_fn(hsize, files)
         hist = getattr(XSH, "history", None)
@@ -256,7 +260,7 @@ class JsonHistoryFlusher(threading.Thread):
         self, filename, buffer, queue, cond, at_exit=False, skip=None, *args, **kwargs
     ):
         """Thread for flushing history."""
-        super(JsonHistoryFlusher, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.filename = filename
         self.buffer = buffer
         self.queue = queue
@@ -299,7 +303,7 @@ class JsonHistoryFlusher(threading.Thread):
 
             cmds.append(cmd)
             last_inp = cmd["inp"]
-        with open(self.filename, "r", newline="\n") as f:
+        with open(self.filename, newline="\n") as f:
             hist = xlj.LazyJSON(f).load()
         load_hist_len = len(hist["cmds"])
         hist["cmds"].extend(cmds)
@@ -360,7 +364,7 @@ class JsonCommandField(cabc.Sequence):
         queue.append(self)
         with self.hist._cond:
             self.hist._cond.wait_for(self.i_am_at_the_front)
-            with open(self.hist.filename, "r", newline="\n") as f:
+            with open(self.hist.filename, newline="\n") as f:
                 lj = xlj.LazyJSON(f, reopen=False)
                 rtn = lj["cmds"][key].get(self.field, self.default)
                 if isinstance(rtn, xlj.LJNode):
@@ -411,9 +415,7 @@ class JsonHistory(History):
         if filename is None:
             # pylint: disable=no-member
             data_dir = _xhj_get_data_dir()
-            self.filename = os.path.join(
-                data_dir, "xonsh-{0}.json".format(self.sessionid)
-            )
+            self.filename = os.path.join(data_dir, f"xonsh-{self.sessionid}.json")
         else:
             self.filename = filename
 
@@ -470,7 +472,7 @@ class JsonHistory(History):
         hf : JsonHistoryFlusher or None
             The thread that was spawned to flush history
         """
-        if not self.remember_history:
+        if (not self.remember_history) or self.is_ignored(cmd):
             return
 
         opts = XSH.env.get("HISTCONTROL", "")
@@ -552,7 +554,7 @@ class JsonHistory(History):
                 continue
             try:
                 commands = json_file.load()["cmds"]
-            except (json.decoder.JSONDecodeError, ValueError):
+            except (JSONDecodeError, ValueError):
                 # file is corrupted somehow
                 if XSH.env.get("XONSH_DEBUG") > 0:
                     msg = "xonsh history file {0!r} is not valid JSON"
@@ -599,3 +601,43 @@ class JsonHistory(History):
 
         # Flush empty history object to disk, overwriting previous data.
         self.flush()
+
+    def delete(self, pattern):
+        """Deletes all entries in history which matches a pattern."""
+        pattern = re.compile(pattern)
+
+        deleted = 0
+        # First, delete any matching commands in the in-memory buffer.
+        for i, cmd in enumerate(self.buffer):
+            if pattern.match(cmd["inp"]):
+                del self.buffer[i]
+                deleted += 1
+
+        # Then, delete any matching commands on disk.
+        while self.gc and self.gc.is_alive():
+            time.sleep(0.011)  # gc sleeps for 0.01 secs, sleep a beat longer
+        for f in _xhj_get_history_files():
+            try:
+                json_file = xlj.LazyJSON(f, reopen=False)
+            except ValueError:
+                # Invalid json file
+                continue
+            try:
+                file_content = json_file.load()
+                commands = file_content["cmds"]
+                for i, c in enumerate(commands):
+                    if pattern.match(c["inp"]):
+                        del commands[i]
+                        deleted += 1
+
+                file_content["cmds"] = commands
+                with open(f, "w") as fp:
+                    xlj.ljdump(file_content, fp)
+            except (JSONDecodeError, ValueError):
+                # file is corrupted somehow
+                if XSH.env.get("XONSH_DEBUG") > 0:
+                    msg = "xonsh history file {0!r} is not valid JSON"
+                    print(msg.format(f), file=sys.stderr)
+                continue
+
+        return deleted
